/*
	ppu.cpp
	fuboy picture processing unit
	/mic 2009
*/

#include "ppu.h"
#include "memory.h"

tag_ppu ppu;
vu8 *dp,*cp,*dp2;
vu16 *dp16;
unsigned char c,sx,sy;
unsigned short mapadr,pal16[4];
unsigned char tnu,pal[4],opal0[4],opal1[4],*opal;


bool ppu_reset() {
	ppu.mode = 0;
	ppu.scanline = 0;
	REG_LCDC = 0x91;
	REG_LY = 0;
	return true;
}




void ppu_draw_sprites(int prio) {
	int i,u,v;
	int spriteNum=0;

	if (REG_LCDC & 2) {
		dp = &frameBuffer[320*ppu.scanline + 0x200];

		// 8x16 OBJ
		if (REG_LCDC & 4) {
			for (i=0; i<40; i++) {
				sy = OAM[i*4] - 16;
				v = 16;
				if ((sy<=ppu.scanline)&&(sy+15>=ppu.scanline))
					v = ppu.scanline - sy;
				if (v<16) {
					sx = OAM[i*4+1] - 8;
					if (((OAM[i*4+3]&0x80)==prio) &&
						((sx<160))) {
						tnu = OAM[i*4+2]&0xFE;
						dp2 = dp + sx;
						if (OAM[i*4+3]&0x40)
							v ^= 15;
						cp = &VRAM[tnu*16 + v*2];
						opal = (OAM[i*4+3]&0x10)?(&opal1[0]):(&opal0[0]);
						if (OAM[i*4+3]&0x20) {
							for (u=7; u>=0; u--) {
								c = ((cp[0]>>(7-u))&1) + ((cp[1]>>(7-u))&1)*2;
								if (c!=0)
									{*dp2 = opal[c]; *(dp2+1) = opal[c]; }
								dp2+=2;
							}
						} else {
							for (u=0; u<8; u++) {
								c = ((cp[0]>>(7-u))&1) + ((cp[1]>>(7-u))&1)*2;
								if (c!=0)
									{*dp2 = opal[c]; *(dp2+1) = opal[c]; }
								dp2+=2;
							}
						}
					}
				}
			}

		// 8x8 OBJ
		} else {
			for (i=0; i<40; i++) {
				sy = OAM[i*4] - 16;
				v = 8;
				if ((sy<=ppu.scanline)&&(sy+7>=ppu.scanline))
					v = ppu.scanline - sy;
				if (v<8) {
					sx = OAM[i*4+1] - 8;
					if (((OAM[i*4+3]&0x80)==prio) &&
						((sx<160))) {
						if (OAM[i*4+3]&0x40)
							v ^= 7;
						tnu = OAM[i*4+2];
						dp2 = dp + sx;
						cp = &VRAM[tnu*16 + v*2];
						opal = (OAM[i*4+3]&0x10)?(&opal1[0]):(&opal0[0]);
						if (OAM[i*4+3]&0x20) {
							for (u=7; u>=0; u--) {
								c = ((cp[0]>>(7-u))&1) + ((cp[1]>>(7-u))&1)*2;
								if (c!=0)
									{*dp2 = opal[c]; *(dp2+1) = opal[c]; }
								dp2+=2;
							}
						} else {
							for (u=0; u<8; u++) {
								c = ((cp[0]>>(7-u))&1) + ((cp[1]>>(7-u))&1)*2;
								if (c!=0)
									{*dp2 = opal[c]; *(dp2+1) = opal[c]; }
								dp2+=2;
							}
						}
					}
				}
			}
		}
	}
}


void ppu_draw_scanline() {
	int i,x,y;
	unsigned short filadr;

	mapadr = (REG_LCDC & 8)?0x1C00:0x1800;
	mapadr += (((ppu.scanline+REG_SCY)&0xFF)>>3)<<5;
	dp = &frameBuffer[320*ppu.scanline + 0x200];
	dp16 = dp;
	y = (ppu.scanline+REG_SCY) & 7;

	pal[0] = REG_BGP&3;
	pal[1] = (REG_BGP>>2)&3;
	pal[2] = (REG_BGP>>4)&3;
	pal[3] = (REG_BGP>>6)&3;

	pal16[0] = pal[0] + (pal[0]<<8);
	pal16[1] = pal[1] + (pal[1]<<8);
	pal16[2] = pal[2] + (pal[2]<<8);
	pal16[3] = pal[3] + (pal[3]<<8);

	opal0[0] = REG_OBP0&3;
	opal0[1] = (REG_OBP0>>2)&3;
	opal0[2] = (REG_OBP0>>4)&3;
	opal0[3] = (REG_OBP0>>6)&3;

	opal1[0] = REG_OBP1&3;
	opal1[1] = (REG_OBP1>>2)&3;
	opal1[2] = (REG_OBP1>>4)&3;
	opal1[3] = (REG_OBP1>>6)&3;

	if ((REG_LCDC & 0x80)==0) return;

	ppu_draw_sprites(0x80);

	// Draw BG
	if (REG_LCDC & 1) {
		tnu = VRAM[mapadr + ((REG_SCX>>3)&0x1F)];
		cp = &VRAM[(tnu<<4)+y+y];
		for (x=(REG_SCX&7);x<8;x++)
		{
			c = ((cp[0]>>(7-x))&1) + (((cp[1]>>(7-x))&1)<<1);
			*dp16 = pal16[c];
			dp16++;
		}

		for (i=1; i<19; i++)
		{
			tnu = VRAM[mapadr + ((((i<<3)+REG_SCX)>>3)&0x1F)];
			if (REG_LCDC & 0x10)
			{
				cp = &VRAM[(tnu<<4) + y + y];
			} else
			{
				if (tnu&0x80)
					cp = &VRAM[0x800 + ((tnu-0x80)<<4) + y + y];
				else
					cp = &VRAM[0x1000 + (tnu<<4) + y + y];
			}
			for (x=0;x<8;x++)
			{
				c = ((cp[0]>>(7-x))&1) + (((cp[1]>>(7-x))&1)<<1);
				*dp16 = pal16[c];
				dp16++;
			}
		}

		tnu = VRAM[mapadr + (((REG_SCX+152)>>3)&0x1F)];
		cp = &VRAM[(tnu<<4)+y+y];
		for (x=0; x<8-(REG_SCX&7);x++)
		{
			c = ((cp[0]>>(7-x))&1) + (((cp[1]>>(7-x))&1)<<1);
			*dp16 = pal16[c];
			dp16++;
		}
	}

	//return;

	// Draw window
	if ((REG_LCDC & 0x20) && (ppu.scanline>=REG_WY)) {
		mapadr = (REG_LCDC & 0x40)?0x1C00:0x1800;
		mapadr += (((ppu.scanline-REG_WY)&0xFF)>>3)<<5;
		dp = &frameBuffer[320*ppu.scanline + 0x200 + REG_WX - 6];
		dp16 = dp;
		y = (ppu.scanline-REG_WY) & 7;

		if (REG_LCDC & 0x10) {	// Tiles at 0x8000-0x8FFF
			for (i=REG_WX-6; i<160; i++) {
				tnu = VRAM[mapadr + (((i-REG_WX+6)>>3)&0x1F)];
				x = (i-REG_WX+6)&7;
				cp = &VRAM[tnu*16 + y*2];
				c = ((cp[0]>>(7-x))&1) + ((cp[1]>>(7-x))&1)*2;
				*dp16 = pal16[c];
				dp16++;
			}

		} else {				// Tiles at 0x8800-0x97FF
			for (i=REG_WX-6; i<160; i++) {
				tnu = VRAM[mapadr + (((i-REG_WX+6)>>3)&0x1F)];
				x = (i-REG_WX+6)&7;
				if (tnu&0x80)
					cp = &VRAM[0x800 + (tnu-0x80)*16 + y*2];
				else
					cp = &VRAM[0x1000 + tnu*16 + y*2];
				c = ((cp[0]>>(7-x))&1) + ((cp[1]>>(7-x))&1)*2;
				*dp16 = pal16[c];
				dp16++;
			}
		}
	}

	ppu_draw_sprites(0x00);
}


void ppu_mode(int mode) {
	if ((mode==0) && (ppu.mode!=0)) {
		if (ppu.scanline < 144)
			ppu_draw_scanline();
		ppu.scanline++;
		if (ppu.scanline == 144) {
			mode = 1;
			REG_IF |= 1;
			if ((IME!=0)&&((REG_IE&1)!=0)) {
				REG_IF &= ~1;
				cpu.halted = false;
				cpu_rst(0x40);
				IME = 0;
			}
		} else if (ppu.scanline == 152) {
			mode = 2;
		} else if (ppu.scanline == 153) {
			mode = 3;
		} else if (ppu.scanline == 154) {
			ppu.scanline = 0;
		}

		if ((mode==0) && (REG_STAT & 8)) {
			//REG_IF |= 2;
			if ((IME!=0)&&((REG_IE&2)!=0)) {
				/*REG_IF &= ~2;
				cpu.halted = false;
				cpu_rst(0x48);
				IME = 0;*/
			}
		}

		REG_LY = ppu.scanline;
		if ((REG_LY == REG_LYC) && (REG_STAT & 0x40)){
			REG_IF |= 2;
			REG_STAT |= 4;
			if ((IME!=0)&&((REG_IE&2)!=0)) {
				REG_IF &= ~2;
				cpu.halted = false;
				cpu_rst(0x48);
				IME = 0;
			}
		} else {
			REG_STAT &= ~4;
		}
	}

	if ((REG_STAT & 0x10) && ((mode==1)||((REG_LCDC&0x80)==0))) {
		//REG_IF |= 2;
		if ((IME!=0)&&((REG_IE&2)!=0)) {
			/*REG_IF &= ~2;
			cpu.halted = false;
			cpu_rst(0x48);
			IME = 0;*/
		}
	}
	if ((REG_STAT & 0x20) && (mode==2)) {
		//REG_IF |= 2;
		if ((IME!=0)&&((REG_IE&2)!=0)) {
			/*REG_IF &= ~2;
			cpu.halted = false;
			cpu_rst(0x48);
			IME = 0;*/
		}
	}

	ppu.mode = mode;
	REG_STAT = (REG_STAT&0xFC)|mode;

}



